﻿# Author: Anthony Stahler
# Last updated: July 24, 2024
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net/resources/
#
# Version 2: added options for multiple file locations, changed from audio normalization to detecting volume and amplifying (easier to troubleshoot matches)
#
# The primary purpose of this script is to detect how much the audio and video are out of sync after you're finished encoding your final video.
# I've noticed that converting DTS mkv files into PCM mov files introduces a lot of audio shift, which is a problem because using something like
# DTS->AC3 as the source audio for Premiere means you're doing multiple lossy encodes, which is suboptimal.  So I wanted a way to objectively see 
# how much of a shift was being introduced so I can compensate for it in my high-quality final encode workflow.
#
#
# This script will give you output that looks something like this:
#    New Timestamp: 00:18:25.60500
#    New Timestamp: 00:12:30.33292
#    A-TOP for The.Gentlemen.1080p.bluray.dd51.mkv,fileorig1c-normalized.raw : v:249 pos:1450059 time:00:00:30.20956
#    V-PTS for The.Gentlemen.1080p.bluray.dd51.mkv : 10.552
#    A-TOP for The.Gentlemen.1080p.bestedit.ver12.mp4,fileorig1c-normalized.raw : v:246 pos:1550439 time:00:00:32.30081
#    V-PTS for The.Gentlemen.1080p.bestedit.ver12.mp4 : 12.6376
#    SECONDS OFFSET (negative = delay audio to reach sync, positive = hasten audio): 0.0056499999999966540
#
# The information you want is primarily the last value, in this case 5.65 milliseconds of hastening that should be added to the audio to make the A/V sync of the edit file
# match the A/V sync of the original MKV file from MakeMKV. (ffmpeg atrim = hasten, adelay = delay, or use VLC's Track Synchronization to test, although it doesn't allow
# sub-millisecond values so it will only be an approximation)  If this value is not in the low milliseconds range, something likely went wrong and you need to adjust your 
# parameters.  You can either pick a new set of timestamps to compare, if the A-TOP v values are not similar, or change your $scenedetect percentage, if the V-PTS values
# are very different from each other.
#
# The V-PTS value is the time in seconds after the offset you chose until ffmpeg detects a visual scene change.  So if you gave it approximately the same starting scene 
# timestamp for both videos, then the offset to detect the change should be similar.  If you get a -1 value it isn't finding a scene at all and you need to lower the 
# percentage until it finds one on both files.  I find 0.30 works pretty well for daytime scenes and 0.20 for night or darker scenes.  But if you lower it too much it may 
# find changes too easily and not be a hard scene change that's a unique data point, resulting in randomly matching different spots in each file since the edit will be 
# encoded slightly different from the original.
#
# The A-TOP value is the peak audio value (one sample) out of the 60 seconds it's scanning for a match.  The main risk is that your audio start and stop times aren't 
# exactly the same, so there's a peak in the start or end of one audio clip that isn't available in the other video.  If this happens - with the detection being too close 
# to the start or end, you can move your timestamps slightly so that part of the audio is outside the matching timeframe. 
#
# It's important to note the New Timestamp values the script gives you - because it doesn't use the exact timestamp you give it.  This is for precision reasons - ffmpeg
# shifting isn't precise if you give it a random time.  The time you give it needs to be at the start of a keyframe, so that the audio samples are exactly aligned with
# the video frame, since we're looking for sub-millisecond precision.  So we search backward from the requested timestamp to find the closest keyframe and use that time.
#
# TIPS FOR USAGE:
# 1. Find a section of your edit with 3+ minutes of uninterupted footage/audio without edits, longer is better since it gives you more leeway to move around
# 2. Locate a unique part of the footage in both videos so you get timestamps close to each other.  They don't have to be exact but try to be within a couple seconds
# 3. Test run the script to see what values it finds - if the final offset is bigger than 5-10ms you probably have a bad match
# 4. If you have a bad match, look at the A-TOP values first, since the audio match isn't very customizable. If the timestamps are more than a couple seconds difference, 
#    then you probably have a hard to match section of audio, so add 1 minute to each of your timestamps and try again.
# 5. Once you have two approximately matching A-TOP values, you can play around with the $scenedetect1/2 values to get a good video match.  It's pretty unlikely that the
#    entire 1 minute segment is a single continuous shot, but if it is obviously you can't use it and need to find a new target timestamp.  Assuming there are shot changes
#    you should be able to get the video to detect it by changing the values.  For example:
#
#    Attempt 1 (bad offset, 14.59 seconds is obviously wrong)
#        New Timestamp: 01:39:39.89100
#        New Timestamp: 01:22:24.64804
#        A-TOP for Quantum.of.Solace.2008.1080p.bluray.dd51.mkv,fileorig1c-normalized.raw : v:255 pos:1682963 time:00:00:35.06173
#        V-PTS for Quantum.of.Solace.2008.1080p.bluray.dd51.mkv : 19.185
#        A-TOP for James.Bond.2006-2015.1080p.bestedit.ver6.mp4,fileorig1c-normalized.raw : v:255 pos:1732996 time:00:00:36.10408
#        V-PTS for James.Bond.2006-2015.1080p.bestedit.ver6.mp4 : 5.63063
#        SECONDS OFFSET (negative = delay audio to reach sync, positive = hasten audio): 14.5967241666666649989
#
#    So since my audio timestamps are fairly close as I expected, I just need to change my $scenedetect1 value down, or $scenedetect2 value up so that they hit the match at
#    the same point.  In this case both were set at 0.30, so I chose to lower $scenedetect1 to 0.20 and leave $scenedetect2 at 0.30, this will make $scenedetect1 more
#    sensitive so it will match sooner.  And here are the results...
#
#    Attempt 2
#        New Timestamp: 01:39:39.89100
#        New Timestamp: 01:22:24.64804
#        A-TOP for Quantum.of.Solace.2008.1080p.bluray.dd51.mkv,fileorig1c-normalized.raw : v:255 pos:1682963 time:00:00:35.06173
#        V-PTS for Quantum.of.Solace.2008.1080p.bluray.dd51.mkv : 4.588
#        A-TOP for James.Bond.2006-2015.1080p.bestedit.ver6.mp4,fileorig1c-normalized.raw : v:255 pos:1732996 time:00:00:36.10408
#        V-PTS for James.Bond.2006-2015.1080p.bestedit.ver6.mp4 : 5.63063
#        SECONDS OFFSET (negative = delay audio to reach sync, positive = hasten audio): -0.0002758333333350011
#



function ConvertSamplesToTime
{
    # this is doing to give us a timestamp with milliseconds to two extra digits of precision, so hundreths of milliseconds
    param($samplespos)

    $timestring = ""
    $samplespos = [decimal]$samplespos # important, things get wonky if you don't explicitly cast this - the following multiply concatenates rather than multiplies.

    # multiply frames by 24fps (1000/24) = 41.66666666666667 to get milliseconds
    [decimal]$totalmillis = $samplespos * (100000.0/$samplesPerSec)

    # round down, we'll remove the remainder in a second
    $tmpmillis = [math]::Floor($totalmillis / 360000000)
    $timestring += $tmpmillis.toString("00") + ":"
    $totalmillis -= ($tmpmillis * 360000000)

    $tmpmillis = [math]::Floor($totalmillis / 6000000)
    $timestring += $tmpmillis.toString("00") + ":"
    $totalmillis -= ($tmpmillis * 6000000)

    $tmpmillis = [math]::Floor($totalmillis / 100000)
    $timestring += $tmpmillis.toString("00") + "."
    $totalmillis -= ($tmpmillis * 100000)

    # add the last milliseconds on the end
    $timestring += $totalmillis.toString("00000")

    return [string]$timestring
}

function ConvertSamplesToDecimal
{
    # this is doing to give us a timestamp with milliseconds to two extra digits of precision, so hundreths of milliseconds
    param($samplespos)

    $timestring = ""
    $samplespos = [decimal]$samplespos # important, things get wonky if you don't explicitly cast this - the following multiply concatenates rather than multiplies.

    # multiply frames by 24fps (1000/24) = 41.66666666666667 to get milliseconds
    [decimal]$totalseconds = $samplespos * (1/$samplesPerSec)

    return $totalseconds
}

function ConvertTimestampToSeconds
{
    param($timestring)

    # takes formats from:
    # 00:00:27.65733
    # 00:27.65733
    # 27.65733
    #

    [decimal]$totalseconds = 0
    $tsarr = $timestring.Trim().Split(':')

    switch ($tsarr.Length)
    {
        1 {
            # just the seconds
            $totalseconds = [System.Convert]::ToDecimal($tsarr[0])
            break
        }

        2 {
            # seconds and minutes
            $totalseconds = [System.Convert]::ToDecimal($tsarr[1])
            $totalseconds += ([System.Convert]::ToInt64($tsarr[0]) * 60) # minutes
            break
        }

        3 {
            # seconds minutes and hours
            $totalseconds = [System.Convert]::ToDecimal($tsarr[2])
            $totalseconds += ([System.Convert]::ToInt64($tsarr[1]) * 60) # minutes
            $totalseconds += ([System.Convert]::ToInt64($tsarr[0]) * 3600) # hours
            break
        }
    }

    return [decimal]$totalseconds
}

function ConvertSecondsToTimestamp
{
    param($totalseconds)

    $timestring = ""
    $totalseconds = [decimal]$totalseconds # important, things get wonky if you don't explicitly cast this - the following multiply concatenates rather than multiplies.

    # round down, we'll remove the remainder in a second
    $tmpseconds = [math]::Floor($totalseconds / 3600)
    $timestring += $tmpseconds.toString("00") + ":"
    $totalseconds -= ($tmpseconds * 3600)

    $tmpseconds = [math]::Floor($totalseconds / 60)
    $timestring += $tmpseconds.toString("00") + ":"
    $totalseconds -= ($tmpseconds * 60)

    $timestring += $totalseconds.toString("00.00000")

    return [string]$timestring
}

# --------------------------
# VARIABLES 
# --------------------------

# uncomment the optional line if you want to use two different file locations
$srcpath1 = "C:\testsync\"
$srcpath2 = $srcpath1
#$srcpath2 = "\\192.168.1.50\test-sync\"


# This script expects two video files in the source folder.
# You need to give it a start offset for each so that it knows what 30 second segment to test for video alignment and audio sync
$fileorig1 = "Battleship.2012.1080p.bluray.dd51.mkv"
$fileedit2 = "Battleship.2012.1080p.bestedit.ver13.mp4"

# these offsets correspond to the names above, and are given in ffmpeg timestamp format
# 9:15 = 9 minutes 15 seconds
# 1:15:20 = 1 hour 15 minutes, 20 seconds
# you don't need to be ultra precise, just make sure both timestamps have the same first scene-change, but if they're a couple seconds off within the same scene that's fine.
# Sometimes these timestamps will be the same for both files, if you're checking at the start of an edit that uses the same beginning as the original,
# but as you get deeper into an edit the times will diverge, so you need to tell the script approximately where to look to match up the video.
#
# Important: make sure the timestamp you choose has at least a minute and a half of no edits after it so the audio can match correctly.
#

$fileorigoffset1 = "3:08"
$fileeditoffset2 = "3:08" 

# typically your audio is going to be 48000 but if not then adjust this
$samplesPerSec = 48000

# by default these will use the same value, but you can set a custom value for each file separately if you need.  This is for situations where one file is padded with 
# black bars and the other isn't, which makes a significant difference, so you might want the one with black bars at 20% and the other at 30% to get a match on the 
# same frame of video.  This is because black bars make the video area bigger, so you'll get less percentage movement for the same picture, so the detection needs to be lower.

# It seems like 0.20 or 0.30 is usually a good value depending on the scene (lower for darker scenes), but sometimes 0.10 or 0.40 are necessary in unique situations.
# Move lower if you aren't getting a match, move higher if you're getting bad matches that aren't scene changes.
$scenedetect1 = 0.30
$scenedetect2 = $scenedetect1
#$scenedetect2 = 0.40

# --------------------------
# --------------------------

$testpaths = @($srcpath1, $srcpath2)
$testfiles = @($fileorig1, $fileedit2)
$testoffsets = @($fileorigoffset1, $fileeditoffset2)
$testdetects = @($scenedetect1, $scenedetect2)
$frametext = "diffframedata.txt"

# I was having issues with inconsistency, and tried having ffmpeg split out the 30 second clip beforehand to reduce any misalignment of video frames versus audio timestamps
# but that didn't work, even though the results were 11ms+ different in alignment on each of the video files, resulting in even bigger gaps and obviously incorrect data.
# So what I want to do instead is find the keyframe timestamp nearest to the chosen offset, and then use that since we should be able to pull a precise alignment of video
# and audio from the start of a keyframe... theoretically.
#
# So I'm going to rewind about a minute from the chosen timestamp and then go forward to the last keyframe before the chosen timestamp, then use that instead.
#
for ($z = 0;$z -lt $testfiles.Length;$z++)
{
    $tmpfile = $testfiles[$z]
    $tmpoffset = $testoffsets[$z]
    $srcpath = $testpaths[$z]
    
    $tmpoffsetseconds = (ConvertTimestampToSeconds $tmpoffset)
    $tmptargetseconds = $tmpoffsetseconds

    $tmpoffsetseconds -= 55 # subtract 55 seconds and use a 60 second scan window, so most of the keyframe scan is prior with a slight overlap of the desired timecode

    if ($tmpoffsetseconds -lt 0)
    {
        $tmpoffsetseconds = 0
    }

    # ffprobe -v 0 -select_streams v -of compact=p=0 -skip_frame nokey -show_frames -show_entries frame=pts_time -read_intervals 500%+60 Drops.of.God.2023.S01E01-08.1080p.bestedit.ver3.mp4
    # -v 0 hides all the file stuff except the pts output we actually want, it's a verbosity level
    $silenceoutput = &"ffprobe" "-v" "0" "-select_streams" "v" "-of" "compact=p=0" "-skip_frame" "nokey" "-show_frames" "-show_entries" "frame=pts_time" "-read_intervals" "$tmpoffsetseconds%+60" "`"$srcpath$tmpfile`"" >$srcpath$z$frametext

    $wholefile = [System.IO.File]::ReadAllText($srcpath + $z + $frametext)

    # standardize newlines
    $wholefile = $wholefile -replace "`r`n", "`n"
    $wholefile = $wholefile -replace "`r", "`n"

    $wholefile_lines = $wholefile -split "`n"

    $highestkeyframe = -1

    ForEach ($linedata_orig in $wholefile_lines) {

        $tmpint = $linedata_orig.IndexOf("pts_time=")

        if ($tmpint -ge 0)
        {
            # expected format:
            # pts_time=57.474083
            # for some reason it sometimes also does this garbage, depending on the file format:
            # pts_time=390.849000|tag:timecode=01:06:29:11|side_data_type=SMPTE 12-1 timecode|value=01:06:29:11
            
            $tmpstr = $linedata_orig.Substring($tmpint + 9)
            $tmpint = $tmpstr.IndexOf("|")

            if ($tmpint -ge 0)
            {
                # get rid of garbage that's sometimes on end
                $tmpstr = $tmpstr.Substring(0,$tmpint)
            }
            
            # sometimes ffmpeg returns N/A instead of an actual number, so we have to check for that...
            if ($tmpstr -ne 'N/A')
            {
                $tmpdec = [System.Convert]::ToDecimal($tmpstr)
        
                if ($tmpdec -le $tmptargetseconds)
                {
                    # it isn't past our target, so it's eligible to be considered

                    if ($tmpdec -gt $highestkeyframe)
                    {
                        $highestkeyframe = $tmpdec
                    }
                }

                # > $tmptargetseconds
                # > previous highest
            }
        }
    }

    # Convert back to timestamp format for ease of human review
    if ($highestkeyframe -ge 0)
    {
        $testoffsets[$z] = (ConvertSecondsToTimestamp $highestkeyframe)

        Write-Host "New Timestamp: $($testoffsets[$z])"
    }
    else
    {
        Write-Host "WARNING: No V-PTS timestamps found for '$srcpath$tmpfile', results may not be accurate!"
    }

    Remove-Item -LiteralPath ("$srcpath$z$frametext") -Force -Recurse
}


$testaudios = @("fileorig1c-normalized.raw", "fileorig1l-normalized.raw")
$iscenter = $true
$testaudiovals = @(0, 0)
$testvideovals = @(0, 0)
$videotext = "diffscenechange.txt"
$normtext = "loudnormout.txt"

for ($z = 0;$z -lt $testfiles.Length;$z++)
{
    $tmpfile = $testfiles[$z]
    $tmpoffset = $testoffsets[$z]
    $srcpath = $testpaths[$z]

    # typically we want to save out the center channel, but if it's a stereo or other format that doesn't use center then we're also grabbing the left in a separate file as a backup
    $silenceoutput = &"ffmpeg" "-y" "-ss" "$tmpoffset" "-t" "60" "-i" "`"$srcpath$tmpfile`"" "-filter_complex" "`"[0:a]pan=mono|c0=FC[L1];[L1]volumedetect`"" "-f" "u8" "-c:a" "pcm_u8" "-ar" "$samplesPerSec" "-" 2>$srcpath$z$normtext

    $wholefile = [System.IO.File]::ReadAllText($srcpath + $z + $normtext)

    # standardize newlines
    $wholefile = $wholefile -replace "`r`n", "`n"
    $wholefile = $wholefile -replace "`r", "`n"

    $wholefile_lines = $wholefile -split "`n"

    $maxvol = ""

    ForEach ($linedata_orig in $wholefile_lines) {
        $tmparr = $linedata_orig.Trim().ToLower().Split(":")
        
        if ($tmparr.Count -eq 2)
        {
            switch -regex ($tmparr[0].Trim())
            {
                ".* max_volume" {
                    $maxvol = ($tmparr[1] -replace '[-a-zA-Z]', ' ').Trim()
                    break
                }

            }
        }
    }
    Remove-Item -LiteralPath ("$srcpath$z$normtext") -Force -Recurse

    $silenceoutput = &"ffmpeg" "-y" "-ss" "$tmpoffset" "-t" "60" "-i" "`"$srcpath$tmpfile`"" "-filter_complex" "`"[0:a]pan=mono|c0=FC[L1];[L1]volume=$maxvol`dB`"" "-f" "u8" "-c:a" "pcm_u8" "-ar" "$samplesPerSec" "`"$srcpath$z$($testaudios[0])`"" 2>$null
    $silenceoutput = &"ffmpeg" "-y" "-ss" "$tmpoffset" "-t" "60" "-i" "`"$srcpath$tmpfile`"" "-filter_complex" "`"[0:a]pan=mono|c0=FL[L1];[L1]volumedetect`"" "-f" "u8" "-c:a" "pcm_u8" "-ar" "$samplesPerSec" "-" 2>$srcpath$z$normtext
    
    
    $wholefile = [System.IO.File]::ReadAllText($srcpath + $z + $normtext)

    # standardize newlines
    $wholefile = $wholefile -replace "`r`n", "`n"
    $wholefile = $wholefile -replace "`r", "`n"

    $wholefile_lines = $wholefile -split "`n"

    $maxvol = ""

    ForEach ($linedata_orig in $wholefile_lines) {
        $tmparr = $linedata_orig.Trim().ToLower().Split(":")
        
        if ($tmparr.Count -eq 2)
        {
            switch -regex ($tmparr[0].Trim())
            {
                ".* max_volume" {
                    $maxvol = ($tmparr[1] -replace '[-a-zA-Z]', ' ').Trim()
                    break
                }

            }
        }
    }
    Remove-Item -LiteralPath ("$srcpath$z$normtext") -Force -Recurse

    $silenceoutput = &"ffmpeg" "-y" "-ss" "$tmpoffset" "-t" "60" "-i" "`"$srcpath$tmpfile`"" "-filter_complex" "`"[0:a]pan=mono|c0=FL[L1];[L1]volume=$maxvol`dB`"" "-f" "u8" "-c:a" "pcm_u8" "-ar" "$samplesPerSec" "`"$srcpath$z$($testaudios[1])`"" 2>$null


    # now analyze these raw files to find the peak value, then we're going to do the same with the edit file
    ForEach ($tmpaudio in $testaudios) {

        $topvalue = -1
        $topcount = -1

        [byte[]] $audioBytes = [System.IO.File]::ReadAllBytes($srcpath + $z + $tmpaudio)

        for ($i=0;$i -lt $audioBytes.Length;$i++)
        {
            $tmpval = [int]$audioBytes[$i]

            if ($tmpval -gt $topvalue)
            {
                $topcount = $i
                $topvalue = $tmpval
            }
        }

        $outtime = ConvertSamplesToTime $topcount

        Write-Host "A-TOP for $tmpfile,$tmpaudio : v:$topvalue pos:$topcount time:$outtime"

        # if it's the first loop and we're on the center channel, mark the channel to use for the subsequent loop
        if ($z -eq 0 -and $tmpaudio -eq $testaudios[0])
        {
            if ($topvalue -gt 128)
            {
                # 128 is 0x80 and is the value of a blank channel
                $iscenter = $true
            }
            else
            {
                $iscenter = $false
            }
        }

        $testaudiovals[$z] = $topcount

        if (($z -eq 0 -and $topvalue -gt 128) -or ($z -gt 0 -and $iscenter))
        {
            # dump out of the loop if the center channel had what we need
            break;
        }
    }

    Remove-Item -LiteralPath ("$srcpath$z$($testaudios[0])") -Force -Recurse
    Remove-Item -LiteralPath ("$srcpath$z$($testaudios[1])") -Force -Recurse
    

    # now that we've done the audio we need to check the video 
    # ffmpeg -ss 3:45 -t 30 -i "%i" -filter:v "select='gt(scene,0.2)',showinfo" -f null - 2> diffscenechange.txt
    $silenceoutput = &"ffmpeg" "-ss" "$tmpoffset" "-t" "30" "-i" "`"$srcpath$tmpfile`"" "-filter:v" "`"select='gt(scene,$($testdetects[$z]))',showinfo`"" "-f" "null" "-" 2>$srcpath$z$videotext
    #Start-Process -NoNewWindow -Wait "C:\Users\User\Documents\Computer\FfMPEG\bin\ffmpeg.exe" -ArgumentList "-loglevel error -hide_banner -nostats -y -itsoffset 0.2 -i $videofilename -i $videofilename -map 0:0 -map 1:1 -vf yadif -c:v:0 libx264 -ar 44100 -scodec copy $newmp4filename" -RedirectStandardError "C:\Users\User\Documents\Computer\FfMPEG\ffreport.txt"
    #Start-Process -NoNewWindow -Wait "ffmpeg" -ArgumentList "-report -ss $tmpoffset -t 30 -i `"$srcpath$tmpfile`" -filter:v `"select='gt(scene,$($testdetects[$z]))',showinfo`" -f null -" -RedirectStandardError "$srcpath$z$videotext"

    # grab the value from the output file
    $scenetmpfile = [System.IO.File]::ReadAllText($srcpath + $z + $videotext)

    # standardize newlines
    $scenetmpfile = $scenetmpfile -replace "`r`n", "`n"
    $scenetmpfile = $scenetmpfile -replace "`r", "`n"

    # search for the value we need
    # [Parsed_showinfo_1 @ 000001c8e3bf6440] n:   0 pts:  98493 pts_time:4.10387 duration:   1001 duration_time:0.0417083 pos:189631031 fmt:yuv422p sar:1/1 s:1920x800 
    $tmppts = -1
    
     #$tmppts = $scenetmpfile -replace '(Parsed_showinfo)([^`n]*)(pts_time:)([\.0-9]+)', "`$4"
    if ($scenetmpfile -match '(.*?)(Parsed_showinfo)(.*?)(pts_time:)([\.0-9]+)(.*)') 
    {
        $tmppts = [decimal]$Matches.5
    }
    
    $testvideovals[$z] = $tmppts
    Write-Host "V-PTS for $tmpfile : $tmppts"

    #$scenetmpfile_lines = $scenetmpfile -split "`n"
    #ForEach ($linedata_orig in $scenetmpfile_lines) {
        
    #}

    Remove-Item -LiteralPath ("$srcpath$z$videotext") -Force -Recurse
}

# now reconcile the values if we have everything
if ($testaudiovals[0] -gt 0 -and $testaudiovals[1] -gt 0 -and $testvideovals[0] -gt 0 -and $testvideovals[1] -gt 0)
{
    # first find the video offset between the two clips
    $tmpvidoffset = $testvideovals[0] - $testvideovals[1]

    #find number of samples difference
    $tmpaudiooffset = (ConvertSamplesToDecimal $testaudiovals[0]) - (ConvertSamplesToDecimal $testaudiovals[1])

    $totaloffset =  $tmpvidoffset - $tmpaudiooffset

    Write-Host "SECONDS OFFSET (negative = delay audio to reach sync, positive = hasten audio): $totaloffset"
}
else
{
    Write-Host "MISSING VALUES, CAN'T SYNC"
}
